////////////////////////////////////////////////////////////////////////////////
// AReflection Library
// Copyright (c) 2004 by Arne Adams
// Permission to use, copy, modify, distribute and sell this software for any 
//     purpose is hereby granted without fee, provided that the above copyright 
//     notice appear in all copies and that both that copyright notice and this 
//     permission notice appear in supporting documentation.
// The author makes no representations about the suitability of this software 
//     for any purpose. It is provided "as is" without express or implied 
//     warranty.
////////////////////////////////////////////////////////////////////////////////
#ifndef REFLIB_BASECLASSITER_H_INC_
#define REFLIB_BASECLASSITER_H_INC_

#include "boost/type_traits/is_same.hpp"
#include <boost/pending/ct_if.hpp>
#include <typeinfo>

#include <Loki/Typelist.h>

#include "../qualifiedtype.h"

namespace refl
{
    struct VirtualBaseClassAlreadyHandledTag{};

    struct VirtualBaseClassNotHandledYetTag{};

    template<class ClassList,class WhichBaseClass, bool isVirtual = true> 
    struct GetVBaseClassHandled
    {
        enum {handledPreviously = Loki::TL::IndexOf<ClassList,WhichBaseClass>::
                                                                    value > -1};
        typedef typename boost::ct_if<(handledPreviously != 0),
                                    VirtualBaseClassAlreadyHandledTag,
                                    VirtualBaseClassNotHandledYetTag >::type
        type;
    };

    template<class ClassList,class WhichBaseClass> struct 
    GetVBaseClassHandled<ClassList, WhichBaseClass, false>
    {
        typedef VirtualBaseClassNotHandledYetTag type;
    };

    template<class HandledVBaseClasses, class BaseClass, class DecisionTag> 
    struct AddBaseClassConditionallyImpl;

    template<class HandledVBaseClasses, class BaseClass> 
    struct AddBaseClassConditionallyImpl<HandledVBaseClasses,BaseClass,
                                        VirtualBaseClassAlreadyHandledTag>
    {
        typedef HandledVBaseClasses type;
    };

    template<class HandledVBaseClasses, class BaseClass> 
    struct AddBaseClassConditionallyImpl<HandledVBaseClasses,BaseClass,
                                        VirtualBaseClassNotHandledYetTag>
    {
        typedef Loki::Typelist<BaseClass,HandledVBaseClasses> type;
    };
    // add a virtual BaseClass if and only if it is not already an element of 
    // HandledVBaseClasses
    template<class HandledVBaseClasses, class BaseClass, bool isVirtual> 
    struct AddBaseClassConditionally
    {
        typedef typename GetVBaseClassHandled<HandledVBaseClasses,BaseClass,
                                            isVirtual>::type DecisionTag;
        typedef typename AddBaseClassConditionallyImpl<HandledVBaseClasses,
                                                    BaseClass,DecisionTag>::type
            type;
    };
    // a given class might be a virtual and a non virtual base class 
    // ISO / IEC 14882 10.1 (6)
    // accumulate only the virtual occurences of a class
    template<class HandledVBaseClasses, class BaseClass>
    struct AddBaseClassConditionally<HandledVBaseClasses,BaseClass,false>
    {
        typedef HandledVBaseClasses type;
    };

    // InitialHandledVBaseClassesT is a Typelist of all classes that have been 
    // previously processed Idea: accumulate that list at compiletime and let 
    // the compiler generate no ops for virtual base classes that have already 
    // been processed.
    template<class ReflectedClass, class InitialHandledVBaseClassesT> 
    struct AllBaseClasses
    {
        typedef InitialHandledVBaseClassesT InitialHandledVBaseClasses;
        enum {TotalNumBaseClasses = ReflectedClass::NumBaseClasses};
        template<class CTRecursion, int Num> struct BaseClassVisitor
        {
            typedef typename ReflectedClass::template GetBaseClass<Num> 
                CurrentBaseClassTraits;
            typedef typename CurrentBaseClassTraits::type CurrentBaseClassType;
            // apply const if the context is const
            typedef typename GetQualifiedMemberType<ReflectedClass,
                                        CurrentBaseClassType>::type BaseClass;

            typedef BaseClassVisitor<CTRecursion,Num - 1> Previous;
            typedef typename Previous::HandledVBaseClasses HandledSofar;
            // pass what was done so far to the next vertical recursion
            typedef AllBaseClasses<BaseClass,HandledSofar> 
                    BaseClassesOfBaseClass;
            // add (recursively) the virtual base classes that where handled by 
            // the next vertical recursion
            typedef typename BaseClassesOfBaseClass::template 
                AfterEnd<CTRecursion>::HandledVBaseClasses 
                SofarPlusVBaseClassesOfBaseClass;
            // add BaseClass to the list of classes handled after this step has 
            // been processed, if it will be handled in this step and is a 
            // virtual base class of ReflectedClass that was not handled so far.
            typedef typename 
                AddBaseClassConditionally<SofarPlusVBaseClassesOfBaseClass, 
                                BaseClass, CurrentBaseClassTraits::isVirtual>::
                                type HandledVBaseClasses;
            static void processSingleBaseClass(ReflectedClass& recordData,
                                            CTRecursion& ctRecursion)
            {
                // HandledSofar instead of SofarPlusVBaseClassesOfBaseClass, 
                // because BaseClass cannot be an element of 
                // SofarPlusVBaseClassesOfBaseClass - HandledSofar (as set 
                // difference) because this would imply that
                // BaseClass is a base class of itself (which won't compile - 
                // hence it is safe to assume that BaseClass is never an element
                // of the set SofarPlusVBaseClassesOfBaseClass - HandledSofar).
                typedef typename GetVBaseClassHandled<HandledSofar,BaseClass,
                                        CurrentBaseClassTraits::isVirtual>::type
                                        VBaseClassHandledTag;
    #ifdef _DEBUG
                const char* baseClassName = typeid(BaseClass).name();
                const char* handledSofa = 
                    typeid(SofarPlusVBaseClassesOfBaseClass).name();
                const char* handledHere = typeid(HandledVBaseClasses).name();
    #endif
                processAsBaseClass(recordData,ctRecursion, 
                                VBaseClassHandledTag());
            } 
            // binary overload
            static void processSingleBaseClass(ReflectedClass& lhsRecord,
                                            ReflectedClass& rhsRecord, CTRecursion& ctRecursion)
            {
                // HandledSofar instead of SofarPlusVBaseClassesOfBaseClass, 
                // because BaseClass cannot be an element of 
                // SofarPlusVBaseClassesOfBaseClass - HandledSofar (as set 
                // difference) because this would imply that
                // BaseClass is a base class of itself (which won't compile - 
                // hence it is safe to assume that BaseClass is never an element
                // of the set SofarPlusVBaseClassesOfBaseClass - HandledSofar).
                typedef typename GetVBaseClassHandled<HandledSofar,BaseClass,
                                        CurrentBaseClassTraits::isVirtual>::type
                                        VBaseClassHandledTag;
    #ifdef _DEBUG
                const char* baseClassName = typeid(BaseClass).name();
                const char* handledSofa = 
                    typeid(SofarPlusVBaseClassesOfBaseClass).name();
                const char* handledHere = typeid(HandledVBaseClasses).name();
    #endif
                processAsBaseClass(lhsRecord, rhsRecord, ctRecursion, 
                                VBaseClassHandledTag());
            } 
            // if CTRecursion is inclined to iterate over static members only
            static void processSingleBaseClass(CTRecursion& ctRecursion)
            {
                // HandledSofar instead of SofarPlusVBaseClassesOfBaseClass, 
                // because BaseClass cannot be an element of 
                // SofarPlusVBaseClassesOfBaseClass - HandledSofar (as set 
                // difference) because this would imply that
                // BaseClass is a base class of itself (which won't compile - 
                // hence it is safe to assume that BaseClass is never an element
                // of the set SofarPlusVBaseClassesOfBaseClass - HandledSofar).
                typedef typename GetVBaseClassHandled<HandledSofar,BaseClass,
                                        CurrentBaseClassTraits::isVirtual>::type
                                        VBaseClassHandledTag;
    #ifdef _DEBUG
                const char* baseClassName = typeid(BaseClass).name();
                const char* handledSofa = 
                    typeid(SofarPlusVBaseClassesOfBaseClass).name();
                const char* handledHere = typeid(HandledVBaseClasses).name();
    #endif
                processAsBaseClass(ctRecursion,VBaseClassHandledTag());
            } 
            static void processAsBaseClass(BaseClass& recordData, 
                                    CTRecursion& ctRecursion, 
                                    VirtualBaseClassNotHandledYetTag)    
            {  
                BaseClassesOfBaseClass::doForeach(recordData,ctRecursion);
                ctRecursion(recordData);
            }
            // binary overload
            static void processAsBaseClass(BaseClass& lhsRecord, 
                                    BaseClass& rhsRecord,
                                    CTRecursion& ctRecursion, 
                                    VirtualBaseClassNotHandledYetTag)    
            {  
                BaseClassesOfBaseClass::doForeach(lhsRecord, rhsRecord,ctRecursion);
                ctRecursion(lhsRecord, rhsRecord);
            }
            static void processAsBaseClass(CTRecursion& ctRecursion, 
                                    VirtualBaseClassNotHandledYetTag)    
            {  
                BaseClassesOfBaseClass::doForeach(ctRecursion);
                ctRecursion.template operator()<BaseClass>();
            }
            // more than once is more then one time too often
            static void processAsBaseClass(BaseClass&, CTRecursion&,
                                        VirtualBaseClassAlreadyHandledTag)
            {}
            // binary overload
            static void processAsBaseClass(BaseClass&, BaseClass&, CTRecursion&,
                                        VirtualBaseClassAlreadyHandledTag)
            {}
            static void processAsBaseClass(CTRecursion&,
                                        VirtualBaseClassAlreadyHandledTag)
            {}                          
        };
        // quick hack to facilitate handling when ReflectedClass has no 
        // baseclass.
        template<class CTRecursion> struct BaseClassVisitor<CTRecursion,-1>
        {
            typedef InitialHandledVBaseClasses HandledVBaseClasses;
        };
        template<class CTRecursion, int RestToDo> struct HorizontalIter:
                    BaseClassVisitor<CTRecursion,TotalNumBaseClasses-RestToDo>
        {
            typedef BaseClassVisitor<CTRecursion,TotalNumBaseClasses-RestToDo> Inherited;
            static void process(ReflectedClass& recordData,
                                CTRecursion& ctRecursion)
            {
                Inherited::Inherited::processSingleBaseClass(recordData, ctRecursion);
                HorizontalIter<CTRecursion,RestToDo-1>::process(recordData, 
                                                            ctRecursion);
            }
            static void process(ReflectedClass& lhsRecord,
                                ReflectedClass& rhsRecord,
                                CTRecursion& ctRecursion)
            {
                Inherited::processSingleBaseClass(lhsRecord, rhsRecord, ctRecursion);
                HorizontalIter<CTRecursion,RestToDo-1>::process(lhsRecord,
                                                            rhsRecord,
                                                            ctRecursion);
            }
            static void process(CTRecursion& ctRecursion)
            {
                Inherited::processSingleBaseClass(ctRecursion);
                HorizontalIter<CTRecursion,RestToDo-1>::process(ctRecursion);
            }
        };

        template<class CTRecursion> struct HorizontalIter<CTRecursion,1>:
                            BaseClassVisitor<CTRecursion,TotalNumBaseClasses-1>
        {
            typedef BaseClassVisitor<CTRecursion,TotalNumBaseClasses-1> Inherited;
            static void process(ReflectedClass& recordData,
                                CTRecursion& ctRecursion)
            {
                Inherited::processSingleBaseClass(recordData, ctRecursion);
            }
            static void process(ReflectedClass& lhsRecord,
                                ReflectedClass& rhsRecord,
                                CTRecursion& ctRecursion)
            {
                Inherited::processSingleBaseClass(lhsRecord, rhsRecord, ctRecursion);
            }
            static void process(CTRecursion& ctRecursion)
            {
                Inherited::processSingleBaseClass(ctRecursion);
            }
        };
        template<class CTRecursion> struct HorizontalIter<CTRecursion,0>
        {
            typedef BaseClassVisitor<CTRecursion,TotalNumBaseClasses-1> Prev;
            typedef typename Prev::HandledVBaseClasses HandledVBaseClasses;
            // RestToDo == 0 process will be called when 
            // TotalNumBaseClasses == 0 as well.
            static void process(ReflectedClass& ,CTRecursion& )
            {}
            // binary overload
            static void process(ReflectedClass& ,ReflectedClass& ,CTRecursion& )
            {}
            static void process(CTRecursion& )
            {}
        };
        // Convenience interface
        template<class CTRecursion> struct AfterEnd
        {
            typedef HorizontalIter<CTRecursion,0> Impl;
            typedef typename Impl::HandledVBaseClasses HandledVBaseClasses;
        };

        template<class CTRecursion> static void
        doForeach(ReflectedClass& dataTuple, CTRecursion& ctRecursion)
        {
            HorizontalIter<CTRecursion,TotalNumBaseClasses>::process(dataTuple,
                                                                ctRecursion);
        }
        // binary overload
        template<class CTRecursion> static void
        doForeach(ReflectedClass& lhsRecord, ReflectedClass& rhsRecord, 
            CTRecursion& ctRecursion)
        {
            HorizontalIter<CTRecursion,TotalNumBaseClasses>::process(lhsRecord,
                                                                rhsRecord, ctRecursion);
        }
        // This form can only access static members of base classes
        // but on the upside is that you dont need an instance of ReflectedClass
        template<class CTRecursion> static void 
        doForeach(CTRecursion& ctRecursion)
        {
            HorizontalIter<CTRecursion,
                         TotalNumBaseClasses>::process(ctRecursion);
        }
    };
    // might as well be useful to have a foreachBaseClassAndThisClass algorithm
    /*template<class ReflectedClass, class CompileTimeRecursion> 
    void foreachBaseClassAndThisClass(ReflectedClass& dataTuple,
                                    CompileTimeRecursion& ctRecursion)   
    {
        typedef typename Loki::TL::MakeTypelist<>::Result EmptyList;
        AllBaseClasses<ReflectedClass,EmptyList>::doForeach(dataTuple,
                                                            ctRecursion);
        ctRecursion(dataTuple);
    }*/
    template<class ReflectedClass, class CompileTimeRecursion> 
    void foreachBaseClass(ReflectedClass& dataTuple, 
                        CompileTimeRecursion& ctRecursion)   
    {
        typedef typename Loki::TL::MakeTypelist<>::Result EmptyList;
        AllBaseClasses<ReflectedClass,EmptyList>::doForeach(dataTuple,
                                                            ctRecursion);
    }
    // binary overload
    template<class ReflectedClass, class CompileTimeRecursion> 
    void foreachBaseClass(ReflectedClass& lhsRecord, ReflectedClass& rhsRecord, 
                        CompileTimeRecursion& ctRecursion)   
    {
        typedef typename Loki::TL::MakeTypelist<>::Result EmptyList;
        AllBaseClasses<ReflectedClass,EmptyList>::doForeach(lhsRecord,
                                                       rhsRecord, ctRecursion);
    }
    template<class ReflectedClass, class CompileTimeRecursion> 
    void foreachBaseClass(CompileTimeRecursion& ctRecursion)   
    {
        typedef typename Loki::TL::MakeTypelist<>::Result EmptyList;
        AllBaseClasses<ReflectedClass,EmptyList>::doForeach(ctRecursion);
    }
}

#endif
