////////////////////////////////////////////////////////////////////////////////
// AReflection Library
// Copyright (c) 2004 by Arne Adams
// Permission to use, copy, modify, distribute and sell this software for any 
//     purpose is hereby granted without fee, provided that the above copyright 
//     notice appear in all copies and that both that copyright notice and this 
//     permission notice appear in supporting documentation.
// The author makes no representations about the suitability of this software 
//     for any purpose. It is provided "as is" without express or implied 
//     warranty.
////////////////////////////////////////////////////////////////////////////////
#ifndef REFLIB_BASECLASSITER_H_INC_
#define REFLIB_BASECLASSITER_H_INC_


#include <typeinfo>
#include <Loki/Typelist.h>

#include "../qualifiedtype.h"

#include <boost/pending/ct_if.hpp>
#include "boost/type_traits/is_same.hpp"

#include "passtype.h"

namespace refl
{
    struct VirtualBaseClassAlreadyHandledTag{};

    struct VirtualBaseClassNotHandledYetTag{};

    template<class ClassList,class WhichBaseClass, bool isVirtual> 
    struct GetVBaseClassHandled
    {
        enum {handledPreviously = Loki::TL::IndexOf<ClassList,WhichBaseClass>::
                                                                    value > -1};
        typedef typename boost::ct_if<(isVirtual ),
                        typename boost::ct_if<(handledPreviously != 0),
                                    VirtualBaseClassAlreadyHandledTag,
                                    VirtualBaseClassNotHandledYetTag >::type,
                        VirtualBaseClassNotHandledYetTag >::type
        type;
    };


    template<class HandledVBaseClasses, class BaseClass, class DecisionTag> 
    struct AddBaseClassConditionallyImpl
    {
        typedef typename boost::ct_if<(boost::is_same<DecisionTag,        
                                VirtualBaseClassAlreadyHandledTag>::value != 0),
                        HandledVBaseClasses,
                        Loki::Typelist<BaseClass,HandledVBaseClasses> >::type
            type;
    };

    // add a virtual BaseClass if and only if it is not already an element of 
    // HandledVBaseClasses
    // a given class might be a virtual and a non virtual base class 
    // ISO / IEC 14882 10.1 (6)
    // accumulate only the virtual occurences of a class
    template<class HandledVBaseClasses, class BaseClass, bool isVirtual> 
    struct AddBaseClassConditionally
    {
        typedef typename GetVBaseClassHandled<HandledVBaseClasses,BaseClass,
                                            isVirtual>::type DecisionTag;
        typedef typename AddBaseClassConditionallyImpl<HandledVBaseClasses,
                                                    BaseClass,DecisionTag>::type
            addedtype;
        typedef typename boost::ct_if<isVirtual,
                                    addedtype,
                                    HandledVBaseClasses>::type
            type;
    };
    namespace detail
    {
        // This is as ugly as can be - MSVC++ 6.0 instantiates typedefs at 
        // template-class scope without having seen the template arguments of 
        // that template class. The default type it uses here is "int"
        typedef int DefaultType;
        template<class T> struct RedirectedAccess
        {
            typedef typename T::type type;
            enum {isVirtual = T::isVirtual};
        };
        template<> struct RedirectedAccess<DefaultType>
        {
            typedef DefaultType type;
            enum {isVirtual = -1};
        };
        // This one as well is only needed to survive the "instantiation"
        // without having seen the template parameters.
        template<class T> struct RedirectedAccess2
        {
            typedef typename T::HandledVBaseClasses HandledVBaseClasses;
        };
        template<class ReflectedClass, class InitialHandledVBaseClassesT, 
                class CTRecursion>
        struct CompileTimeRecursionLog
        {
            typedef InitialHandledVBaseClassesT InitialHandledVBaseClasses;
            enum {TotalNumBaseClasses = ReflectedClass::NumBaseClasses};
            template<int Num> struct HorizontalIterImpl
            {
                typedef typename ReflectedClass::BaseClasses BaseClassList;
                typedef typename Loki::TL::TypeAt<BaseClassList, Num>::Result 
                    CurrentBaseClassTraits;
                // typedef typename CurrentBaseClassTraits::type 
                // CurrentBaseClassType;
                typedef typename RedirectedAccess<CurrentBaseClassTraits>::type
                    CurrentBaseClassType;
                // apply const if the context is const
                typedef typename GetQualifiedMemberType<ReflectedClass,
                                            CurrentBaseClassType>::type 
                                            BaseClass;

                typedef HorizontalIterImpl<Num - 1> Previous;
                typedef typename Previous::HandledVBaseClasses HandledSofar;
                // pass what was done so far to the next vertical recursion
                typedef CompileTimeRecursionLog<BaseClass,HandledSofar,
                                                CTRecursion> 
                        BaseClassesOfBaseClass;
                // add (recursively) the virtual base classes that where handled
                // by the next vertical recursion
                /*typedef typename BaseClassesOfBaseClass::HandledVBaseClasses 
                    SofarPlusVBaseClassesOfBaseClass;*/
                typedef typename 
                RedirectedAccess2<BaseClassesOfBaseClass>::HandledVBaseClasses 
                    SofarPlusVBaseClassesOfBaseClass;
                // add BaseClass to the list of classes handled after this step 
                // has been processed, if it will be handled in this step and is
                // a virtual base class of ReflectedClass that was not handled 
                // so far.
                typedef typename 
                AddBaseClassConditionally<SofarPlusVBaseClassesOfBaseClass,
                                    BaseClass, 
                (bool)RedirectedAccess<CurrentBaseClassTraits>::isVirtual>::type
                HandledVBaseClasses;

                // HandledSofar instead of SofarPlusVBaseClassesOfBaseClass, 
                // because BaseClass cannot be an element of 
                // SofarPlusVBaseClassesOfBaseClass - HandledSofar (as set 
                // difference) because this would imply that
                // BaseClass is a base class of itself (which won't compile - 
                // hence it is safe to assume that BaseClass is never an element
                // of the set SofarPlusVBaseClassesOfBaseClass - HandledSofar).
                typedef typename GetVBaseClassHandled<HandledSofar,BaseClass,
                (bool)RedirectedAccess<CurrentBaseClassTraits>::isVirtual>::type
                VBaseClassHandledTag;
            };
            template<> struct HorizontalIterImpl<-1>
            {
                typedef InitialHandledVBaseClasses HandledVBaseClasses;
            };
            // Convenience interface
            typedef typename 
                HorizontalIterImpl<TotalNumBaseClasses-1>::HandledVBaseClasses 
                HandledVBaseClasses;
        };
    }
    // InitialHandledVBaseClassesT is a Typelist of all classes that have been 
    // previously processed Idea: accumulate that list at compiletime and let 
    // the compiler generate no ops for virtual base classes that have already 
    // been processed.
    template<class ReflectedClass, class InitialHandledVBaseClassesT> 
    struct AllBaseClasses
    {
        typedef InitialHandledVBaseClassesT InitialHandledVBaseClasses;
        enum {TotalNumBaseClasses = ReflectedClass::NumBaseClasses};
        template<int Num> struct BaseClassVisitor
        {
            template<class CTRecursion> static void processSingleBaseClass(
                                            ReflectedClass& recordData,
                                            CTRecursion& ctRecursion)
            {

                typedef detail::CompileTimeRecursionLog<ReflectedClass, 
                                        InitialHandledVBaseClassesT, 
                                        CTRecursion> RecursionControl;
                typedef typename RecursionControl::template 
                    HorizontalIterImpl<Num> CompileTimeLogEntry;
                typedef typename CompileTimeLogEntry::BaseClass BaseClass;
                typedef typename CompileTimeLogEntry::VBaseClassHandledTag 
                    VBaseClassHandledTag;
    #ifdef _DEBUG
                const char* baseClassName = typeid(BaseClass).name();
    #endif
                // this cast from derived to base is necessary because this way 
                // I can pass the type of the baseclass to processAsBaseClass 
                processAsBaseClass(static_cast<BaseClass&>(recordData),
                                ctRecursion, 
                                VBaseClassHandledTag());
            } 
            // binary overload
            template<class CTRecursion> static void 
                processSingleBaseClass(ReflectedClass& lhsRecord,ReflectedClass& rhsRecord,
                                       CTRecursion& ctRecursion)
            {

                typedef detail::CompileTimeRecursionLog<ReflectedClass, 
                                        InitialHandledVBaseClassesT, 
                                        CTRecursion> RecursionControl;
                typedef typename RecursionControl::template 
                    HorizontalIterImpl<Num> CompileTimeLogEntry;
                typedef typename CompileTimeLogEntry::BaseClass BaseClass;
                typedef typename CompileTimeLogEntry::VBaseClassHandledTag 
                    VBaseClassHandledTag;
    #ifdef _DEBUG
                const char* baseClassName = typeid(BaseClass).name();
    #endif
                // this cast from derived to base is necessary because this way 
                // I can pass the type of the baseclass to processAsBaseClass 
                processAsBaseClass(static_cast<BaseClass&>(lhsRecord),
                                static_cast<BaseClass&>(rhsRecord),
                                ctRecursion, 
                                VBaseClassHandledTag());
            } 
            // if CTRecursion is inclined to iterate over static members only
            template<class CTRecursion> static void processSingleBaseClass(
                                                     CTRecursion& ctRecursion)
            {
                // typedefs at class scope that refer a dependent member 
                // template cause compiler errors with VC++ 6.0
                typedef detail::CompileTimeRecursionLog<ReflectedClass, 
                                        InitialHandledVBaseClassesT, 
                                        CTRecursion> RecursionControl;
                typedef typename 
                    RecursionControl::template HorizontalIterImpl<Num> 
                    CompileTimeLogEntry;
                typedef typename CompileTimeLogEntry::BaseClass BaseClass;
                typedef typename CompileTimeLogEntry::VBaseClassHandledTag 
                    VBaseClassHandledTag;
    #ifdef _DEBUG
                const char* baseClassName = typeid(BaseClass).name();
    #endif
                processAsBaseClass(ctRecursion,VBaseClassHandledTag(),
                                PassType<BaseClass>());
            } 
            // there can only be one type of baseclass in this instantiation of 
            // BaseClassVisitor - unfortunately that type is not available at 
            // class scope because MSVC++ 6.0 tries to instantiate the needed 
            // typedefs without having seen the template arguments and then 
            // chokes.
            template<class BaseClass, class CTRecursion> static void 
            processAsBaseClass(BaseClass& recordData, 
                              CTRecursion& ctRecursion, 
                              VirtualBaseClassNotHandledYetTag)    
            {  
                typedef detail::CompileTimeRecursionLog<ReflectedClass, 
                                        InitialHandledVBaseClassesT, 
                                        CTRecursion> RecursionControl;
                typedef typename 
                    RecursionControl::template HorizontalIterImpl<Num> 
                    CompileTimeLogEntry;
                typedef typename CompileTimeLogEntry::HandledSofar HandledSofar;
                typedef AllBaseClasses<BaseClass,HandledSofar> 
                    BaseClassesOfBaseClass;
                BaseClassesOfBaseClass::doForeach(recordData,ctRecursion);
                ctRecursion(recordData);
            }
            // binary overload
            template<class BaseClass, class CTRecursion> static void 
            processAsBaseClass(BaseClass& lhsRecord, 
                              BaseClass& rhsRecord,
                              CTRecursion& ctRecursion, 
                              VirtualBaseClassNotHandledYetTag)    
            {  
                typedef detail::CompileTimeRecursionLog<ReflectedClass, 
                                        InitialHandledVBaseClassesT, 
                                        CTRecursion> RecursionControl;
                typedef typename 
                    RecursionControl::template HorizontalIterImpl<Num> 
                    CompileTimeLogEntry;
                typedef typename CompileTimeLogEntry::HandledSofar HandledSofar;
                typedef AllBaseClasses<BaseClass,HandledSofar> 
                    BaseClassesOfBaseClass;
                BaseClassesOfBaseClass::doForeach(lhsRecord, rhsRecord,ctRecursion);
                ctRecursion(lhsRecord, rhsRecord);
            }
            template<class BaseClass,class CTRecursion> static void 
            processAsBaseClass(CTRecursion& ctRecursion, 
                        VirtualBaseClassNotHandledYetTag,PassType<BaseClass>)
            {  
                typedef detail::CompileTimeRecursionLog<ReflectedClass, 
                                        InitialHandledVBaseClassesT, 
                                        CTRecursion> RecursionControl;
                typedef typename 
                    RecursionControl::template HorizontalIterImpl<Num> 
                    CompileTimeLogEntry;
                typedef typename CompileTimeLogEntry::HandledSofar HandledSofar;
                typedef AllBaseClasses<BaseClass,HandledSofar> 
                    BaseClassesOfBaseClass;
                BaseClassesOfBaseClass::doForeach(ctRecursion);
                ctRecursion(PassType<BaseClass>(),ClassVariableTag());
            }
            // more than once is more then one time too often
            template<class BaseClass, class CTRecursion> static void 
            processAsBaseClass(BaseClass&, CTRecursion&,
                               VirtualBaseClassAlreadyHandledTag)
            {}
            // binary overload
            template<class BaseClass, class CTRecursion> static void 
            processAsBaseClass(BaseClass&, BaseClass&, CTRecursion&,
                               VirtualBaseClassAlreadyHandledTag)
            {}
            // more than once is more then one time too often
            template<class BaseClass, class CTRecursion> static void 
            processAsBaseClass(CTRecursion&,
                        VirtualBaseClassAlreadyHandledTag,PassType<BaseClass>)
            {}                          
        };
        template<int RestToDo> struct HorizontalIter:
                    BaseClassVisitor<TotalNumBaseClasses-RestToDo>
        {
            template<class CTRecursion> static void 
            process(ReflectedClass& recordData,CTRecursion& ctRecursion)
            {
                processSingleBaseClass(recordData, ctRecursion);
                HorizontalIter<RestToDo-1>::process(recordData, 
                                                    ctRecursion);
            }
            template<class CTRecursion> static void 
            process(ReflectedClass& lhsRecord,ReflectedClass& rhsRecord,
                CTRecursion& ctRecursion)
            {
                processSingleBaseClass(lhsRecord, rhsRecord, ctRecursion);
                HorizontalIter<RestToDo-1>::process(lhsRecord, 
                                              rhsRecord, ctRecursion);
            }
            template<class CTRecursion> static void 
            process(CTRecursion& ctRecursion)
            {
                processSingleBaseClass(ctRecursion);
                HorizontalIter<RestToDo-1>::process(ctRecursion);
            }
        };

        template<> struct HorizontalIter<1>:
                            BaseClassVisitor<TotalNumBaseClasses-1>
        {
            template<class CTRecursion> static void 
            process(ReflectedClass& recordData,CTRecursion& ctRecursion)
            {
                processSingleBaseClass(recordData, ctRecursion);
            }
            template<class CTRecursion> static void 
            process(ReflectedClass& lhsRecord,ReflectedClass& rhsRecord,
                CTRecursion& ctRecursion)
            {
                processSingleBaseClass(lhsRecord,rhsRecord, ctRecursion);
            }
            template<class CTRecursion> static void 
            process(CTRecursion& ctRecursion)
            {
                processSingleBaseClass(ctRecursion);
            }
        };
        template<> struct HorizontalIter<0>
        {
            // RestToDo == 0 process will be called when 
            // TotalNumBaseClasses == 0 as well.
            template<class CTRecursion> static void 
            process(ReflectedClass& ,CTRecursion& )
            {}
            template<class CTRecursion> static void 
            process(ReflectedClass& ,ReflectedClass& ,CTRecursion& )
            {}
            template<class CTRecursion> static void process(CTRecursion& )
            {}
        };
        template<class CTRecursion> static void
        doForeach(ReflectedClass& dataTuple, CTRecursion& ctRecursion)
        {
            HorizontalIter<TotalNumBaseClasses>::process(dataTuple,
                                                        ctRecursion);
        }
        template<class CTRecursion> static void
        doForeach(ReflectedClass& lhsRecord,ReflectedClass& rhsRecord, CTRecursion& ctRecursion)
        {
            HorizontalIter<TotalNumBaseClasses>::process(lhsRecord,
                                                        rhsRecord, ctRecursion);
        }
        // This form can only access static members of base classes
        // but on the upside is that you dont need an instance of ReflectedClass
        template<class CTRecursion> static void 
        doForeach(CTRecursion& ctRecursion)
        {
            HorizontalIter<TotalNumBaseClasses>::process(ctRecursion);
        }
    };
    // might as well be useful to have a foreachBaseClassAndThisClass algorithm
    /*template<class ReflectedClass, class CompileTimeRecursion> 
    void foreachBaseClassAndThisClass(ReflectedClass& dataTuple,
                                    CompileTimeRecursion& ctRecursion)   
    {
        typedef typename Loki::TL::MakeTypelist<>::Result EmptyList;
        AllBaseClasses<ReflectedClass,EmptyList>::doForeach(dataTuple,
                                                            ctRecursion);
        ctRecursion(dataTuple);
    }*/
    template<class ReflectedClass, class CompileTimeRecursion> 
    void foreachBaseClass(ReflectedClass& dataTuple, 
                        CompileTimeRecursion& ctRecursion)   
    {
        typedef typename Loki::TL::MakeTypelist<>::Result EmptyList;
        AllBaseClasses<ReflectedClass,EmptyList>::doForeach(dataTuple,
                                                            ctRecursion);
    }
    template<class ReflectedClass, class CompileTimeRecursion> 
    void foreachBaseClass(ReflectedClass& lhsRecord,ReflectedClass& rhsRecord, 
                        CompileTimeRecursion& ctRecursion)   
    {
        typedef typename Loki::TL::MakeTypelist<>::Result EmptyList;
        AllBaseClasses<ReflectedClass,EmptyList>::doForeach(lhsRecord,
                                                       rhsRecord, ctRecursion);
    }
    template<class ReflectedClass, class CompileTimeRecursion> 
    void foreachBaseClass(CompileTimeRecursion& ctRecursion)   
    {
        typedef typename Loki::TL::MakeTypelist<>::Result EmptyList;
        AllBaseClasses<ReflectedClass,EmptyList>::doForeach(ctRecursion);
    }
}

#endif
