////////////////////////////////////////////////////////////////////////////////
// AReflection Library
// Copyright (c) 2004 by Arne Adams
// Permission to use, copy, modify, distribute and sell this software for any 
//     purpose is hereby granted without fee, provided that the above copyright 
//     notice appear in all copies and that both that copyright notice and this 
//     permission notice appear in supporting documentation.
// The author makes no representations about the suitability of this software 
//     for any purpose. It is provided "as is" without express or implied 
//     warranty.
////////////////////////////////////////////////////////////////////////////////
#ifndef REFLIB_CONDITIONAL_LOCK_H_INC_
#define REFLIB_CONDITIONAL_LOCK_H_INC_

#include "../userconfig.h"

#include "../qualifiedtype.h"

namespace refl
{
    namespace detail
    {
        template<class ConstNess> struct WrappedLock
        {
            template<class ReflectedClass> struct Impl
            {
                // the default ctor uses the class lock (this is usesful for locking
                // write access to static member variables
                typedef userconfig::MutexInstance<ReflectedClass,
                                    userconfig::ThreadingModel> ReflectedClassMutex;
                typedef typename userconfig::LockType<ReflectedClass,
                                                    userconfig::ThreadingModel>::type
                    type;
                Impl(ReflectedClass& host):lock_(host.getMutex())
                {}
                Impl():lock_(ReflectedClassMutex::get())
                {}
                type lock_;
            };
        };
        template<> struct WrappedLock<TypeIsConstTag>
        {
            template<class ReflectedClass> struct Impl
            {
                Impl()
                {}
                Impl(ReflectedClass& )
                {}
            };
        };
    }
    // lock only write access that is if ReflectedClass is not const
    template<class ReflectedClass, 
        class ConstNess = IsConst<ReflectedClass>::type> struct ConditionalLock
    {
        typedef typename detail::WrappedLock<ConstNess>::template Impl<ReflectedClass> Lock;
        explicit ConditionalLock(ReflectedClass& host):lock_(host)
        {}
        ConditionalLock(){}
    private:
        Lock lock_;
    };
}
#endif
