////////////////////////////////////////////////////////////////////////////////
// AReflection Library
// Copyright (c) 2004 by Arne Adams
// Permission to use, copy, modify, distribute and sell this software for any 
//     purpose is hereby granted without fee, provided that the above copyright 
//     notice appear in all copies and that both that copyright notice and this 
//     permission notice appear in supporting documentation.
// The author makes no representations about the suitability of this software 
//     for any purpose. It is provided "as is" without express or implied 
//     warranty.
////////////////////////////////////////////////////////////////////////////////
#ifndef REFLIB_INITREFLECTEDMEMBERS_H_INC_
#define REFLIB_INITREFLECTEDMEMBERS_H_INC_


#include <boost/type_traits/is_pod.hpp>

#include <boost/pending/ct_if.hpp>

#include "../flatforeach.h"
 
#include "../priviligedaccess.h"

namespace refl
{       
    // default-initializes even numerical and pointer membervariables.
    class RecordInitializer
    {
    // who the hack knows why MSVC++ 6.0 wants these tags public
    public:
        struct DoInitialize{};
        struct DontInitialize{};
        template<class T,int IsRefMember> struct NeedsInit
        {
            enum {isPod = boost::is_pod<T>::value};  
            enum {isConst = boost::is_const<T>::value};  
            typedef typename boost::ct_if<(isPod != 0 && isConst == 0 && 
                                            IsRefMember == 0),
                                   DoInitialize,
                                   DontInitialize >::type
            type;
        };
    public:
        template<class ReflectedClass> void 
        initReflectedMembers(ReflectedClass& initThis)  
        {        
            // note that static members are zero initialized
            // (ISO/IEC 14882 3.6.2 Initialization of nonlocal objects)
            foreachAttribute(initThis,*this,InstanceVariableTag());
        }                                                                 
        template<class ValueType, class NameTag, class WrappedScope> 
        void operator()(ValueType& attribute, NameTag, WrappedScope)
        {
            typedef typename WrappedScope::type Scope;
            typedef typename Scope::template GetValueTypeByTag<NameTag, 
                                                            Scope>::type 
                AttributeValueType;
            enum {isRefMember = 
                ::boost::is_reference<AttributeValueType>::value};
            initValue(attribute, NeedsInit<ValueType,isRefMember>::type());
        }
    private:
        // initialization is only needed for those types that will be otherwise 
        // garbage initialized.
        template<class Value> void
        initValue(Value& out,DoInitialize)
        {                   // this statement is especially valid for built in 
            out = Value();  // types:int out = int();
                            // std::cout << out << std::endl;
        }                   // yields: 0
        
        template<class Value> void initValue(Value& val,DontInitialize)
        { }                               

        /* not more to do because container-elements are default initialized
        by c++ standard.
        Hence if container::value_type is a pod all will be fine and
        if the value_type is an instance of a compile-time iterable
        class then the default ctor of that class was called and hence
        all members of that instance will be default initialized.
        (the latter because the default ctor of a compile-time iterable
        class default initializes all pod-members).*/
    };
    
    template<class ReflectedClass> 
    void initReflectedMembers(ReflectedClass& initThis)
    {
        RecordInitializer initializer;
        initializer.initReflectedMembers(initThis);
    }
}
#endif // !

